# coding=utf-8
__author__ = "Dimitris Karkalousos"

ATOMMIC_ENV_VARNAME_ENABLE_COLORING = "ATOMMIC_ENABLE_COLORING"
ATOMMIC_ENV_VARNAME_REDIRECT_LOGS_TO_STDERR = "ATOMMIC_REDIRECT_LOGS_TO_STDERR"
# Set to True to enable atommic.util.logging's debug mode
ATOMMIC_ENV_VARNAME_TESTING = "ATOMMIC_TESTING"
# Used for atommic.utils.exp_manager versioning
ATOMMIC_ENV_VARNAME_VERSION = "ATOMMIC_EXPM_VERSION"
# Used to change default atommic cache directory
ATOMMIC_ENV_CACHE_DIR = "ATOMMIC_CACHE_DIR"
# Used to change default atommic data store cache directory
ATOMMIC_ENV_DATA_STORE_CACHE_DIR = "ATOMMIC_DATA_STORE_CACHE_DIR"
# Shared among nodes (1) or not shared (0)
ATOMMIC_ENV_DATA_STORE_CACHE_SHARED = "ATOMMIC_DATA_STORE_CACHE_SHARED"
