# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.package_info import (  # noqa: F401
    __contact_emails__,
    __contact_names__,
    __description__,
    __download_url__,
    __homepage__,
    __keywords__,
    __license__,
    __package_name__,
    __repository_url__,
    __shortversion__,
    __version__,
)
