# coding=utf-8

from atommic.collections.multitask import rs  # noqa: F401
from atommic.package_info import __version__

# Set collection version equal to atommic version.
__version = __version__

# Authorship.
__author__ = "Dimitris Karkalousos"

# Set collection name.
__description__ = "MRI multitask learning collection."
