# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.multitask.rs.nn.idslr import IDSLR  # noqa: F401
from atommic.collections.multitask.rs.nn.idslr_unet import IDSLRUNet  # noqa: F401
from atommic.collections.multitask.rs.nn.mtlrs import MTLRS  # noqa: F401
from atommic.collections.multitask.rs.nn.mtlrs_base.mtlrs_block import MTLRSBlock  # noqa: F401
from atommic.collections.multitask.rs.nn.recseg_unet import RecSegUNet  # noqa: F401
from atommic.collections.multitask.rs.nn.segnet import SegNet  # noqa: F401
from atommic.collections.multitask.rs.nn.seranet import SERANet  # noqa: F401
