# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.nn.attentionunet import SegmentationAttentionUNet  # noqa: F401
from atommic.collections.segmentation.nn.dynunet import SegmentationDYNUNet  # noqa: F401
from atommic.collections.segmentation.nn.lambdaunet import SegmentationLambdaUNet  # noqa: F401
from atommic.collections.segmentation.nn.unet import SegmentationUNet  # noqa: F401
from atommic.collections.segmentation.nn.unet3d import Segmentation3DUNet  # noqa: F401
from atommic.collections.segmentation.nn.unetr import SegmentationUNetR  # noqa: F401
from atommic.collections.segmentation.nn.vnet import SegmentationVNet  # noqa: F401
