# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.losses.cross_entropy import CrossEntropyLoss  # noqa: F401
from atommic.collections.segmentation.losses.dice import Dice  # noqa: F401
